from typing import Optional, Protocol

from ...traits.with_stages.with_stages import WithStages
from ...traits.with_builderable.with_builderable import WithBuilderable
from ...traits.with_callable_pipeline.with_callable_pipeline import (
    WithCallablePipeline,
)
from ...traits.with_pipe.with_pipe import WithPipe
from ...traits.with_processor.processor_interface import ProcessorInterface
from ...traits.with_processor.with_processor import WithProcessor
from ...traits.with_immutability.with_immutability import WithImmutability
from .pipeline_callable import PipelineCallable
from .payload import TPayload


class PipelineInterface(
    WithImmutability[TPayload],
    WithCallablePipeline[TPayload],
    WithProcessor[TPayload],
    WithPipe[TPayload],
    WithStages[TPayload],
    WithBuilderable[TPayload],
    Protocol[TPayload],
):
    def __init__(
        self,
        processor: Optional[ProcessorInterface[TPayload]] = None,
        stages: Optional[list[PipelineCallable[TPayload, ...]]] = None,
    ) -> None:
        self.set_processor(processor or self.get_processor())

        self.set_parts([*stages] if stages else self.get_parts())



from typing import Optional, Protocol, Self

from .processor_interface import ProcessorInterface
from ..with_builderable.with_builderable import WithBuilderable
from ..with_stages.with_stages import WithStages
from ...core.pipeline.payload import TPayload


class WithProcessor(
    WithStages[TPayload],
    WithBuilderable[TPayload],
    Protocol[TPayload],
):
    processor_class: Optional[type[ProcessorInterface[TPayload]]] = None

    processor: Optional[ProcessorInterface[TPayload]] = None

    async def process(
        self,
        payload: TPayload,
    ) -> TPayload:
        return await self.get_processor().process(self.get_parts(), payload)

    def get_processor(self) -> ProcessorInterface[TPayload]:
        if self.processor:
            return self.processor

        if self.processor_class is None:
            raise ValueError("No processor set")

        if self.processor is None:
            self.processor = self.processor_class()

        return self.processor

    def set_processor(self, processor: ProcessorInterface[TPayload]) -> Self:
        self.processor = processor

        return self



from typing import Protocol
from thecodecrate_builderable import WithParts

from ...core.pipeline.pipeline_callable import PipelineCallable
from ...core.pipeline.payload import TPayload


class WithBuilderable(
    WithParts[TPayload, PipelineCallable[TPayload, ...]],
    Protocol[TPayload],
):
    pass



from typing import Generic

from .payload import TPayload
from .pipeline_interface import PipelineInterface
from ...modules.processors.chained_processor import ChainedProcessor


class Pipeline(
    PipelineInterface[TPayload],
    Generic[TPayload],
):
    processor_class = ChainedProcessor



from typing import Generic

from ...core.pipeline.pipeline_callable import PipelineCallable
from ...core.pipeline.payload import TPayload
from ...traits.with_processor.processor_interface import ProcessorInterface


class ChainedProcessor(
    ProcessorInterface[TPayload],
    Generic[TPayload],
):
    async def process(
        self,
        stages: list[PipelineCallable[TPayload, ...]],
        payload: TPayload,
    ) -> TPayload:
        for stage in stages:
            payload = await self._call_stage(stage, payload)

        return payload



from typing import Protocol
from thecodecrate_builderable import WithImmutableParts

from ...core.pipeline.payload import TPayload
from ...core.pipeline.pipeline_callable import PipelineCallable


class WithImmutability(
    WithImmutableParts[TPayload, PipelineCallable[TPayload, ...]],
    Protocol[TPayload],
):
    pass



from abc import abstractmethod
import inspect
from typing import Any, Protocol

from ...core.pipeline.payload import TPayload
from ...core.pipeline.pipeline_callable import PipelineCallable


class ProcessorInterface(Protocol[TPayload]):
    @abstractmethod
    async def process(
        self,
        stages: list[PipelineCallable[TPayload, ...]],
        payload: TPayload,
    ) -> TPayload:
        pass

    async def _call_stage(
        self,
        stage: PipelineCallable[TPayload, ...],
        payload: TPayload,
        *args: Any,
        **kwds: Any,
    ) -> TPayload:
        result = stage(payload, *args, **kwds)

        if inspect.isawaitable(result):
            return await result

        return result



from .core.pipeline.pipeline import Pipeline
from .core.pipeline.pipeline_interface import PipelineInterface
from .traits.with_stages.stage_interface import StageInterface
from .traits.with_processor.processor_interface import ProcessorInterface
from .core.pipeline_builder.pipeline_builder import PipelineBuilder
from .core.pipeline_builder.pipeline_builder import PipelineBuilderInterface
from .core.pipeline.pipeline_callable import PipelineCallable
from .modules.processors.chained_processor import ChainedProcessor
from .modules.processors.interruptible_processor import InterruptibleProcessor
from .traits.with_builderable.with_builderable import WithBuilderable
from .traits.with_callable_pipeline.with_callable_pipeline import (
    WithCallablePipeline,
)
from .traits.with_immutability.with_immutability import WithImmutability
from .traits.with_pipe.with_pipe import WithPipe
from .traits.with_pipeline_builder.with_pipeline_builder import (
    WithPipelineBuilder,
)
from .traits.with_processor.with_processor import WithProcessor
from .traits.with_stages.with_stages import WithStages

# Version of the package
# DO NOT MODIFY MANUALLY
# This will be updated by `bumpver` command.
# - Make sure to commit all changes first before running `bumpver`.
# - Run `bumpver update --[minor|major|patch]`
__version__ = "1.6.0"

# Expose the public API
__all__ = [
    # Core
    "Pipeline",
    "PipelineInterface",
    "StageInterface",
    "ProcessorInterface",
    "PipelineBuilder",
    "PipelineBuilderInterface",
    "PipelineCallable",
    # Processors
    "ChainedProcessor",
    "InterruptibleProcessor",
    # Traits
    "WithCallablePipeline",
    "WithBuilderable",
    "WithImmutability",
    "WithPipe",
    "WithPipelineBuilder",
    "WithProcessor",
    "WithStages",
]



from typing import Protocol

from ..with_builderable.with_builderable import WithBuilderable
from ..with_processor.with_processor import WithProcessor
from ..with_stages.with_stages import WithStages
from ..with_stages.stage_interface import StageInterface
from ...core.pipeline.payload import TPayload


class WithCallablePipeline(
    StageInterface[TPayload],
    WithProcessor[TPayload],
    WithStages[TPayload],
    WithBuilderable[TPayload],
    Protocol[TPayload],
):
    async def __call__(self, payload: TPayload) -> TPayload:
        return await self.process(payload)



from abc import abstractmethod
from typing import Any, Protocol
from ...core.pipeline.payload import TPayload


class StageInterface(Protocol[TPayload]):
    @abstractmethod
    async def __call__(
        self,
        payload: TPayload,
        *args: Any,
        **kwds: Any,
    ) -> TPayload:
        pass



from typing import Protocol

from ...traits.with_stages.with_stages import WithStages
from ...traits.with_builderable.with_builderable import WithBuilderable
from ...traits.with_pipeline_builder.with_pipeline_builder import (
    WithPipelineBuilder,
)
from ..pipeline.payload import TPayload


class PipelineBuilderInterface(
    WithPipelineBuilder[TPayload],
    WithStages[TPayload],
    WithBuilderable[TPayload],
    Protocol[TPayload],
):
    def __init__(self) -> None:
        self.set_parts([])



from typing import Protocol, Self

from ..with_builderable.with_builderable import WithBuilderable
from ..with_stages.with_stages import WithStages
from ...core.pipeline.payload import TPayload
from ...core.pipeline.pipeline_callable import PipelineCallable


class WithPipe(
    WithStages[TPayload],
    WithBuilderable[TPayload],
    Protocol[TPayload],
):
    def pipe(self, stage: PipelineCallable[TPayload, ...]) -> Self:
        return self.add_part(stage)



from typing import Generic

from ..pipeline.payload import TPayload
from .pipeline_builder_interface import PipelineBuilderInterface


class PipelineBuilder(
    PipelineBuilderInterface[TPayload],
    Generic[TPayload],
):
    pass



from typing import Protocol, Self

from ..with_builderable.with_builderable import WithBuilderable
from ...core.pipeline.pipeline_callable import PipelineCallable
from ...core.pipeline.payload import TPayload


class WithStages(
    WithBuilderable[TPayload],
    Protocol[TPayload],
):
    stages: list[PipelineCallable[TPayload, ...]] = []

    def get_parts(self) -> list[PipelineCallable[TPayload, ...]]:
        return self.stages

    def set_parts(self, parts: list[PipelineCallable[TPayload, ...]]) -> Self:
        self.stages = parts

        return self



from typing import Optional, Protocol, Self

from ..with_builderable.with_builderable import WithBuilderable
from ..with_processor.with_processor import WithProcessor
from ..with_stages.with_stages import WithStages
from ..with_processor.processor_interface import ProcessorInterface
from ...core.pipeline.pipeline_callable import PipelineCallable
from ...core.pipeline.payload import TPayload
from ...core.pipeline.pipeline import Pipeline


class WithPipelineBuilder(
    WithProcessor[TPayload],
    WithStages[TPayload],
    WithBuilderable[TPayload],
    Protocol[TPayload],
):
    def add(self, part: PipelineCallable[TPayload, ...]) -> Self:
        return self.add_part(part)

    def build(
        self,
        processor: Optional[ProcessorInterface[TPayload]] = None,
    ) -> Pipeline[TPayload]:
        return Pipeline[TPayload](processor=processor, stages=self.get_parts())



from typing import Awaitable, Callable, Concatenate, ParamSpec
from ...core.pipeline.payload import TPayload
from ...traits.with_stages.stage_interface import StageInterface

TArgs = ParamSpec("TArgs")  # To capture *args and **kwds

PipelineCallable = (
    StageInterface[TPayload]
    | Callable[Concatenate[TPayload, TArgs], Awaitable[TPayload]]
    | Callable[Concatenate[TPayload, TArgs], TPayload]
)
